#include "DictionaryTrie.h"

#include <cstdio>

DictionaryTrie::TrieNode::TrieNode() {
	for(unsigned int i = 0; i<26; ++i) {
		childtypes[i] = 0;
	}
}

bool DictionaryTrie::Load(const std::string& file) {
	
	// load file
	std::vector<char> filecontents;
	{
		FILE *f;
		f = fopen(file.c_str(), "rb");
		if(f==NULL) {
			return false;
		}
		try {
			fseek(f, 0, SEEK_END);
			filecontents.resize(ftell(f));
			fseek(f, 0, SEEK_SET);
			unsigned int r = fread(&filecontents.front(), 1, filecontents.size(), f);
			if(r!=filecontents.size()) {
				fclose(f);
				return false;
			}
		} catch(...) {
			fclose(f);
			throw;
		}
		fclose(f);
	}
	
	// convert newlines
	std::string contents;
	{
		for(unsigned int i = 0; i<filecontents.size(); ++i) {
			if(filecontents[i]!='\r') {
				contents.push_back(filecontents[i]);
			}
		}
	}
	filecontents.clear();
	
	// split into words
	{
		unsigned int pos = 0;
		words.clear();
		for( ; ; ) {
			unsigned int end = contents.find('\n', pos);
			if(end==std::string::npos) {
				break;
			}
			words.push_back(contents.substr(pos, end-pos));
			pos = end+1;
		}
		//words.push_back(contents.substr(pos)); // last line
	}
	
	// create trie
	trie.clear();
	trie.push_back(TrieNode());
	for(unsigned int i = 0; i<words.size(); ++i){
		const std::string &word = words[i];
		for(unsigned int j = 0; j<word.size(); ++j) {
			unsigned char mytype = (j==0)? 2 : 1;
			unsigned int current = 0;
			for(unsigned int k = j; k<word.size(); ++k) {
				char c = word[k];
				if(c>='a' && c<='z') {
					unsigned int letter = c-'a';
					unsigned char type = trie[current].childtypes[letter];
					if(type==0) {
						trie[current].childtypes[letter] = mytype;
						trie[current].childs[letter] = trie.size();
						current = trie.size();
						trie.push_back(TrieNode());
					} else {
						if(type==1 && mytype>1) {
							trie[current].childtypes[letter] = mytype;
						}
						current = trie[current].childs[letter];
					}
				} else if(k==j) {
					goto ignoreletter;
				}
			}
			{
				EndNode temp;
				temp.type = mytype;
				temp.index = i;
				trie[current].endnodes.push_back(temp);
			}
			ignoreletter:;
		}
	}
	
	return true;
	
}

std::vector<std::string> DictionaryTrie::Find(const std::string& filter) {
	std::vector<std::string> results;
	
	if(trie.size()==0) {
		return results;
	}
	
	std::string str = filter;
	bool xbegin = false, xend = false;
	if(str.length()!=0) {
		if(str[str.length()-1]=='*') {
			xend = true;
			str = str.substr(0, str.length()-1);
		}
	}
	if(str.length()!=0) {
		if(str[0]=='*') {
			xbegin = true;
			str = str.substr(1);
		}
	}
	
	unsigned char lowesttype = (xbegin)? 1 : 2;
	
	std::vector<unsigned int> currentset;
	currentset.push_back(0);
	for(unsigned int i = 0; i<str.length(); ++i) {
		char c = str[i];
		if(c=='%') {
			std::vector<unsigned int> nextset;
			for(unsigned int j = 0; j<currentset.size(); ++j) {
				unsigned int current = currentset[j];
				for(unsigned int letter = 0; letter<26; ++letter) {
					unsigned char type = trie[current].childtypes[letter];
					if(type>=lowesttype) {
						nextset.push_back(trie[current].childs[letter]);
					}
				}
			}
			currentset.swap(nextset);
		} else if(c>='a' && c<='z') {
			unsigned int letter = c-'a';
			std::vector<unsigned int> nextset;
			for(unsigned int j = 0; j<currentset.size(); ++j) {
				unsigned int current = currentset[j];
				unsigned char type = trie[current].childtypes[letter];
				if(type>=lowesttype) {
					nextset.push_back(trie[current].childs[letter]);
				}
			}
			currentset.swap(nextset);
		}
	}
	if(xbegin && xend && currentset.size()==1 && currentset[0]==0) {
		// the search string was empty or invalid
		lowesttype = 2;
	}
	
	std::vector<unsigned int> tempresults;
	while(currentset.size()!=0) {
		for(unsigned int i = 0; i<currentset.size(); ++i) {
			unsigned int current = currentset[i];
			const std::vector<EndNode> &endnodes = trie[current].endnodes;
			for(unsigned int j = 0; j<endnodes.size(); ++j) {
				if(endnodes[j].type>=lowesttype) {
					tempresults.push_back(endnodes[j].index);
				}
			}
		}
		if(!xend) {
			break;
		}
		std::vector<unsigned int> nextset;
		for(unsigned int i = 0; i<currentset.size(); ++i) {
			unsigned int current = currentset[i];
			for(unsigned int letter = 0; letter<26; ++letter) {
				unsigned char type = trie[current].childtypes[letter];
				if(type>=lowesttype) {
					nextset.push_back(trie[current].childs[letter]);
				}
			}
		}
		currentset.swap(nextset);
	}
	
	// remove duplicates and create final results
	if(tempresults.size()>0) {
		results.push_back(words[tempresults[0]]);
		for(unsigned int i = 1; i<tempresults.size(); ++i) {
			if(tempresults[i]!=tempresults[i-1]) {
				results.push_back(words[tempresults[i]]);
			}
		}
	}
	
	return results;
	
}

